 /*
 	2010/06/09(Wed)
	Minesweeper ver 1.0
	Programmed by KOYAMA Yoshiaki

	Copyright © 2010 KOYAMA Yoshiaki 
	All rights reserved.
*/

#include "iocs.h"
#include "library.h"

/*
	画像データ
*/
byte box[8][10] =  {
			{0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0xff},			/* Box (セル) 枠 OR */
			{0x03, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x03},			
			{0x00, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0x00},				 /* Box (セル) 枠消去	AND	*/
			{0xfc, 0xfd, 0xfd, 0xfd, 0xfd, 0xfd, 0xfd, 0xfd, 0xfd, 0xfc},			
			{0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0xff},			/* 	Box (セル) 内消去	AND*/
			{0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xff},		
			{0x00, 0xfe, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0xfe, 0x00},			/* Box (セル) 1ドット内側枠 */
			{0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00}
		};	

/*
	セルの画像テータ
*/
byte data[15][8] = {
		{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},		/* 0  */		
		{0x00,0x00,0x04,0x02,0x7E,0x00,0x00,0x00},		/*  1 */		
		{0x00,0x00,0x64,0x52,0x4A,0x44,0x00,0x00},		/* 2*/		
		{0x00,0x00,0x22,0x4A,0x4A,0x36,0x00,0x00},		/* 3 */		
		{0x00,0x10,0x18,0x14,0x7E,0x10,0x00,0x00},		/*  4*/		
		{0x00,0x00,0x2E,0x4A,0x4A,0x32,0x00,0x00},		/*  5*/		
		{0x00,0x00,0x3C,0x4A,0x4A,0x32,0x00,0x00},		/*  6*/		
		{0x00,0x00,0x02,0x72,0x0A,0x06,0x00,0x00},		/*  7*/		
		{0x00,0x00,0x34,0x4A,0x4A,0x34,0x00,0x00},		/*  8*/		
		{0x00,0x08,0x3C,0x64,0x26,0x3C,0x10,0x00},		/*  mine */		
		{0x00,0x42,0x24,0x18,0x18,0x24,0x42,0x00},		/*  X */		
		{0x00,0x7E,0x42,0x5A,0x5A,0x42,0x7E,0x00},		/*  block */	
		{0x00,0x00,0x04,0x52,0x0a,0x04,0x00,0x00},		/*  ? */			
		{0x00,0x00,0x7e,0x0e,0x0e,0x0e,0x00,0x00},		/*  flag */	
		{0x00,0x08,0x3C,0x7c,0x3e,0x3c,0x10,0x00}		/* 爆発した地雷*/					
		};

byte gameOver[16] = {' ', ' ', ' ', 'G','a','m','e',' ','O','v','e','r', ' ',' ',' ', 0x00};
byte gameClear[17] = {' ', ' ', ' ', 'G','a','m','e',' ','C','l','e','a','r', ' ',' ',' ', 0x00};

#define BOX_X					15
#define BOX_Y					5
#define BOX_SIZE			(BOX_X*BOX_Y)

#define MINE_NUM			12
/*#define MINE_NUM			20*/

/*
	edata[y][x]
	周りの地雷数 

0xff:	壁
0-8:	地雷の数
9: 		地雷
*/
byte edata[BOX_Y+2][BOX_X+2];

/*
	ｍdata[y][x]
	セル(x,y) が開いているか

0xff:	壁
0:		開いていない
1:		開いた
*/
byte mdata[BOX_Y+2][BOX_X+2];

/*
	cdata[y][x]
	旗が立っているか

0xff:	壁
0:		なし
1:		旗
2:		？
*/
byte cdata[BOX_Y+2][BOX_X+2];

/*
	地雷の数を設定
*/
void setNumber(byte x, byte y) {
	byte d;
	
	d = edata[y][x];
		
	if ((x==0) ||  (y == 0)) {
		return;
	}
	if ((x>=(BOX_X+1)) ||  (y >= (BOX_Y+1))) {
		return;
	}
	
	if ((d != 9) && (d != 0xff)) {
		edata[y][x]++;
	}
}

/*
	セル(x, y) の周りのセルに地雷の数を設定する
*/
void setMine(byte x, byte y)（{
	 edata[y+1][x+1] = 9;
	
	setNumber(x+1, y+1-1);
	setNumber(x+1, y+1+1);
	
	setNumber(x+1-1, y+1-1);
	setNumber(x+1-1, y+1);
	setNumber(x+1-1, y+1+1);
	
	setNumber(x+1+1, y+1-1);
	setNumber(x+1+1, y+1);
	setNumber(x+1+1, y+1+1);	
}

/*
	セル (x,y) の画像を描画
*/
void draw(byte x, byte y, byte drawflag) {
	byte d, m, c;
	
	d = edata[y][x];
	m = mdata[y][x];
	c = cdata[y][x];

	gprint((x-1)*9+4, (y-1)*9, 10,  2, box[4], GPRINT_AND | GPRINT_GRAM);			
			
	if ((m == 0)  && (c == 0)) {				
		gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[11],  GPRINT_OR | drawflag | GPRINT_GRAM);
	} else if ((m == 0) && (c != 0)) {					
		switch (c) {
		  case 1: /* flag */
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[13],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		  case 2: /* ? */
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[12],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		}
	} else {
		switch(d) {		
		 case 0:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[0],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 1:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[1],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 2:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[2],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 3:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[3],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 4:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[4],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 5:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[5],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 6:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[6],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		case 7:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[7],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		case 8:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[8],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		 case 9:
			gprint8((x-1)*9+1+4,(y-1)*9+1, 8, data[9],  GPRINT_OR | drawflag | GPRINT_GRAM);
			break;
		}
	}	
}
/*
	すべて表示

	drawAll(0) の時、メモリに書き込むが表示しない
*/
void drawAll(byte drawflag) {
	byte i, j;
	
	for (i=1; i<=BOX_Y; i++) {
		for (j=1;  j<=BOX_X; j++) {
			draw(j, i, drawflag);
		}
	}
}

/*
		初期化
*/

/*
extern	word _seed;
*/

void init() {
	byte rdata[BOX_SIZE];
	byte i, d, j , k, x, y;
	
/*	_seed = 0x2689;*/
	
	/*
		セル(x,y) を1次元配列 rdata[x*y] に割り当てる
	*/
	for (i=0; i<BOX_SIZE; i++) {
		rdata[i] = i;
	}
	
	/*
		rdata[i]内をシャッフルする。
	*/
	for (i=0; i<(BOX_SIZE  - 1); i++)（{
		k = BOX_SIZE - 1 -  i;
		
		j = random() % k;
		
		d=rdata[k];
		rdata[k]=rdata[j];
		rdata[j] = d;
	}
		
	initgvram();
	
	for (i=0; i<BOX_X; i++) {
		for (j=0; j<BOX_Y; j++) {
			gprint(i*9+4, j*9, 10,  2, box[0], GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);
		}	
	}
	
	for (i=0; i<BOX_Y+2; i++) {
		for (j=0; j<BOX_X+2; j++) {
			if ((i==0) || (i==(BOX_Y+1))) {
				edata[i][j] = 0xff;
				mdata[i][j] = 0xff;
				cdata[i][j] = 0xff;
			} else if ((j == 0)	|| (j==BOX_X+1)) {			
				edata[i][j] = 0xff;								
				mdata[i][j] = 0xff;
				cdata[i][j] = 0xff;				
			} else {
				edata[i][j] = 0;								
				mdata[i][j] = 0;
				cdata[i][j] = 0;			
			}
		}	
	}
	
	/*
		rdata[0] - rdata[MINE_NUM-1] の値が地雷がある場所になる。
	*/
	for (i=0; i<MINE_NUM; i++) {
		y = rdata[i] / BOX_X;
		x = rdata[i] - y*BOX_X;
		
		setMine(x, y);
	}
}

/*
	ゲ−ムをクリアしたか判定

1:	クリア
0:
*/
byte hantei() {
	byte i, j, count, ret;
	
	count = 0;
	
	for (i=1; i<=BOX_Y; i++) {
		for (j=1; j<=BOX_X; j++) {
			if (mdata[i][j]  == 0) {
				count++;
			}
		}
	}
	
	if (count == MINE_NUM) {
		ret = 1;
	} else {
		ret = 0;
	}
	
	return ret;
}

/*
	周りに地雷がない edata[y][x] == 0 の時、連鎖してセルを開く
*/
void rensa(byte x, byte y) {	
	
	if ((edata[y][x] == 0xff) || (mdata[y][x] == 1) || (cdata[y][x] == 1)) {
		return;
	}
	
	if (edata[y][x] == 0) {
		mdata[y][x] = 1;
		rensa(x-1, y-1);
		rensa(x-1, y);
		rensa(x-1, y+1);
		
		rensa(x  ,y-1);
		rensa(x  ,y+1);
		
		rensa(x+1, y-1);
		rensa(x+1, y);
		rensa(x+1, y+1);
	} else {
		mdata[y][x] = 1;
	}
}

/*
	地雷を開く
*/
void openMines() {
	byte i, j;;
	
	for (i=1; i<=BOX_Y; i++) {
		for (j=1;  j<=BOX_X; j++) {
			if( (mdata[i][j] == 0) && (cdata[i][j] != 1)) {			
				if (edata[i][j] == 9) {
					gprint((j-1)*9+4, (i-1)*9, 10,  2, box[4], GPRINT_AND | GPRINT_GRAM);			
					gprint8((j-1)*9+1+4,(i-1)*9+1, 8, data[9],  GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);			
				}
			}
			
			 if (edata[i][j] == 10) {
				gprint((j-1)*9+4, (i-1)*9, 10,  2, box[4], GPRINT_AND | GPRINT_GRAM);			
				gprint8((j-1)*9+1+4,(i-1)*9+1, 8, data[14],  GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);	/* explosion mine */
			}

			if ((mdata[i][j] == 0) && (edata[i][j] != 9) && (cdata[i][j] == 1)) {
					gprint((j-1)*9+4, (i-1)*9, 10,  2, box[4], GPRINT_AND | GPRINT_GRAM);	
					gprint8((j-1)*9+1+4,(i-1)*9+1, 8, data[10],  GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);	 /* batsu */						
			}
		}
	}
	
}

/*
セル(x,y) の周りに立っている旗の数があっているか？

1:	正しい
0: 間違い
*/
byte isMatchRoundFlag(byte x, byte y) {
	byte m, e, c;
	byte mineNum;
	
	mineNum = 0;
	m = mdata[y][x];
	e = edata[y][x];
	c = cdata[y][x];
	
	if (e == 0) {
		return 0;
	}

	if ((mdata[y-1][x-1] == 0) && (cdata[y-1][x-1] == 1)) {
		mineNum++;
	}
	if ((mdata[y][x-1] == 0) && (cdata[y][x-1] == 1)) {
		mineNum++;
	}
	if ((mdata[y+1][x-1] == 0) && (cdata[y+1][x-1] == 1)) {
		mineNum++;
	}
	
	if ((mdata[y-1][x] == 0) && (cdata[y-1][x] == 1)) {
		mineNum++;
	}
	if ((mdata[y+1][x] == 0) && (cdata[y+1][x] == 1)) {
		mineNum++;
	}
	
	if ((mdata[y-1][x+1] == 0) && (cdata[y-1][x+1] == 1)) {
		mineNum++;
	}
	
	if ((mdata[y][x+1] == 0) && (cdata[y][x+1] == 1)) {
		mineNum++;
	}
	
	if ((mdata[y+1][x+1] == 0) && (cdata[y+1][x+1] == 1)) {
		mineNum++;
	}
		
	if (e == mineNum) {
		return 1;
	}
	
	return 0;
}

/*
	スペースキーを押して旗が立っていないセルを自動的に開く

0:
1:	ゲームオーバー
*/
byte roundOpen(byte x, byte y) {
	byte ret;
	
	ret = 0;
	
	if (mdata[y-1][x-1] == 0) {
		if (edata[y-1][x-1] == 9) {
			if (cdata[y-1][x-1] == 1) {
			} else {
				edata[y-1][x-1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y-1][x-1] == 1) {
			} else {
				rensa(x-1, y-1);
			}
		}		
	}
	
	if (mdata[y][x-1] == 0) {
		if (edata[y][x-1] == 9) {
			if (cdata[y][x-1] == 1) {
			} else {
				edata[y][x-1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y][x-1] == 1) {
			} else {
				rensa(x-1, y);
			}
		}
	}
	
	if (mdata[y+1][x-1] == 0) {
		if (edata[y+1][x-1] == 9) {
			if (cdata[y+1][x-1] == 1) {
			} else {
				edata[y+1][x-1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y+1][x-1] == 1) {
			} else {
				rensa(x-1, y+1);
			}
		}		
	}
	
	if (mdata[y-1][x] == 0) {
		if (edata[y-1][x] == 9) {
			if (cdata[y-1][x] == 1) {
			} else {
				edata[y-1][x] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y-1][x] == 1) {
			} else {
				rensa(x, y-1);
			}
		}		
	}
	
	if (mdata[y+1][x] == 0) {
		if (edata[y+1][x] == 9) {
			if (cdata[y+1][x] == 1) {
			} else {
				edata[y+1][x] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y+1][x] == 1) {
			} else {
				rensa(x, y+1);
			}
		}	
	}
	
	if (mdata[y-1][x+1] == 0)  {
		if (edata[y-1][x+1] == 9) {
			if (cdata[y-1][x+1] == 1) {
			} else {
				edata[y-1][x+1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y-1][x+1] == 1) {
			} else {
				rensa(x+1, y-1);
			}
		}				
	}
	
	if (mdata[y][x+1] == 0) {
		if (edata[y][x+1] == 9) {
			if (cdata[y][x+1] == 1) {
			} else {
				edata[y][x+1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y][x+1] == 1) {
			} else {
				rensa(x+1, y);
			}
		}		
	}
	
	if (mdata[y+1][x+1] == 0)  {
		if (edata[y+1][x+1] == 9) {
			if (cdata[y+1][x+1] == 1) {
			} else {
				edata[y+1][x+1] = 10;
				ret = 1;
			}			
		} else {
			if (cdata[y+1][x+1] == 1) {
			} else {
				rensa(x+1, y+1);
			}
		}			
	}
	
	return ret;
}
		
		

/*
	グローバル変数
*/
	byte gcode;								/* キ−コード */
	byte gx, gy, gxx, gyy;			/* セルの位置*/
	byte gend;								/* ゲーム終了フラグ*/
	byte gd, gm, gc;					/*	gd = edata[gy][gx]; gm = mdata[gy][gx]; gc = cdata[gy][gx]; */
	byte gescape ;						/* プログラム終了フラグ */

/*
	変数初期化
*/
void initVariable() {
	gcode = 0;
	gend = 0;
	gx = gy = gxx = gyy = 1;
	gd = gm = gc =0;
}

/*
	ゲームオーバー表示
*/
void doGameOver() {
	setLocate((X_LINE - strlength(gameOver))/2 ,3);
	gprint((gx-1)*9+4, (gy-1)*9, 10,  2, box[0], GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);					
	openMines();
	print(gameOver);	
}

/*
	ゲームクリア表示
*/
void doGameClear() {
	setLocate((X_LINE - strlength(gameClear))/2 ,3);
	gprint((gx-1)*9+4, (gy-1)*9, 10,  2, box[0], GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);	
	print(gameClear);				
}

/*
	ゲーム終了処理

	ゲームが終了した後、何かキーを押すことでゲーム結果を再表示
*/
void doGameEnd() {
	gcode = inkeyWait();
	drawScreen();
	gcode = inkeyWait();
					
	if (gcode == KEY_BREAK) {
		gescape = 1;
	}				
	
	initVariable();
	
	if (gescape == 0) {
		cls();
		init();	
		drawAll(GPRINT_DRAW);	
	}
}

/*
	カーソル位置の計算
*/
void coordinate() {
	gxx = gx;
	gyy = gy;
	
	if ((gcode == KEY_8) || (gcode ==KEY_UP)){
			gy--;
			if (gy<1) {
				gy = 1;
			}
		} else if ((gcode == KEY_2) || (gcode ==KEY_DOWN)) {
			gy++;
			if (gy>BOX_Y) {
				gy = BOX_Y;
			}
		} else if ((gcode == KEY_4) || (gcode ==KEY_LEFT)) {
			gx--;
			if (gx < 1) {
				gx = 1;
			}			
		} else if ((gcode == KEY_6) || (gcode ==KEY_RIGHT)) {
			gx++;
			if (gx > BOX_X) {
				gx = BOX_X ;
			}			
		} 
}

/*
	メイン処理
*/
void main(void) {	
	
	gescape = 0;
	
	initVariable();
	cls();
	init();
	drawAll(GPRINT_DRAW);
	
	/*
		ループ
	*/
	while ((gcode != KEY_BREAK)  && (gescape == 0)){
		
		/* キー入力を待つ */
		gcode = inkeyWait();
		
		coordinate();
		
		gd = edata[gy][gx];
		gm = mdata[gy][gx];
		gc = cdata[gy][gx];
		
		if ((gxx != gx) || (gyy != gy)) {
			gprint((gxx-1)*9+4, (gyy-1)*9, 10,  2, box[0], GPRINT_OR | GPRINT_DRAW | GPRINT_GRAM);	
		}
		gprint((gx-1)*9+4, (gy-1)*9, 10,  2, box[2], GPRINT_AND | GPRINT_DRAW | GPRINT_GRAM);	

		
		if (gcode == KEY_SPACE) {
			/*
				旗が立っていない
			*/
			if (gc != 1) {
				/*
					セルが開いていない
				*/
				if (gm == 0) {
					rensa(gx,gy);
					drawAll(0);
					drawScreen();
			
					if (edata[gy][gx] == 9) {
						edata[gy][gx] = 10;
						doGameOver();
						gend = 1;
					} else if (hantei() == 1) {
						doGameClear();
						gend = 1;
					}
					if (gend == 1) {						
						doGameEnd();
					}
				/*
					セルが開いている
				*/	
				} else { /* m == 1*/
					/*
						セルの周りの旗の数があっているか？
					*/
					if (isMatchRoundFlag(gx,gy) == 1) {
						/*
							地雷を開いた
						*/
						if (roundOpen(gx,gy) == 1) {
							drawAll(0);
							drawScreen();							
							openMines();
							doGameOver();
							doGameEnd();	
						} else {
							drawAll(0);
							drawScreen();
							if (hantei() == 1) {
								doGameClear();
								doGameEnd();
							}	
						}	
					}
				}
			}
		/*
			旗を立てる処理
		*/	
		} else if (gcode == KEY_F) {
			if (gm == 0) {
				switch (gc) {
				 case 0:
					cdata[gy][gx] = 1;
					break;
				 case 1:
					cdata[gy][gx] = 2;
					break;
				 case 2:
					cdata[gy][gx] = 0;
					break;
				default:
					cdata[gy][gx] = 0;
					break;					
				}
			}
			/*
				旗の描画
			*/
			draw(gx, gy, GPRINT_DRAW);
		}
	}
}